/*
 *  
 *  Copyright (c) 2004 Steve Slaven, All Rights Reserved.
 *  
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Incanvas., 59 Temple Place, Suite 330, Boston,
 *  MA 02111-1307 USA
 *  
*/
package net.hoopajoo.cheeselab3;
import java.util.*;
import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;

/**
 * 
 * Experimental threaded map renderer. Should prevent GUI freezes.
 * Based on Stevens's experimental cached rendering engine.
 * 
 * @author Antti Andreimann <Antti.Andreimann@mail.ee>
 */
class CheeseRendererThreaded extends CheeseRendererCached
                             implements Runnable, CheeseMapListener,
                                        CheeseCanvasListener {
    public static final int no_overlays_thr = 5;
    static int threadCounter = 0;
    Thread rendererThread = null;
    VolatileImage osc = null;
    VolatileImage oldosc = null;
    Date last_render = null;

    int oldosc_x = 0, oldosc_y = 0, oldosc_z = 0;
    Object osclock;

    volatile int rapid_re_render_count = 0;
    volatile boolean stopThread = false,
                     waiting = false,
                     complete = false,
                     skiprender = false;
    volatile long needtorender = 0,
                  rendering = 0;
    
    // Collect rendering statistics, globally
    private static volatile long max_time = 0,
                                 avg_time = 0,
                                 min_time = Long.MAX_VALUE,
                                 last_time = 0;

    volatile GraphicsConfiguration gc = null;
    
    // Variables for dealing with graphical room movement
    private Area move_area = null;
    private HashMap move_rooms = new HashMap();
    private HashMap move_affected_rooms = new HashMap();
    private boolean move_repaint = false;
    private Vector move_hidden_rooms = new Vector();
    
    CheeseRendererThreaded() {
        super();
        threadCounter++;
        
        int prio = Thread.MIN_PRIORITY;

        osclock = new Object();

        CheeseLab.dmesg( 2, "Creating map painting thread" );

        rendererThread = new Thread( this, "MapPainterThread-" + threadCounter );

        CheeseLab.dmesg( 3, "Setting thread priority to " + prio);
        try {
            rendererThread.setPriority( prio );
        } catch (Exception e) {
            CheeseLab.dmesg( 2, "Priority setting failed: " + e.getMessage() );
        }
        
        CheeseMap.addCheeseMapListener( this );

        rendererThread.start();
    }

    public void prepare() {
        Dimension tile_size = canvas.get_tile_size();
        if( tile_width != tile_size.width ) {

            synchronized(osclock) {
                clearImage(oldosc);
                clearImage(osc);
            }
        }

        // if( render_grid != canvas.get_render_grid() ) {
        //	memorytiles.clear();
        //}

        // This seems to be very expensive operation when called from
        // non-painting thread so we call it here.
        gc = canvas.getGraphicsConfiguration();

        // Let super update our properties
        super.prepare();
    }

    public void run() {
        CheeseLab.dmesg( 2, rendererThread.getName() + " started");

        while( !stopThread ) {
            while( needtorender == rendering ) {
                // Wait for repaint notification
                synchronized(this) {
                    try {
                        wait();
                    } catch(InterruptedException e) {}
                }
                if( stopThread )
                    return;
            }

            Date rstart = new Date();
            complete = false;

            synchronized(osclock) {

                if( last_render != null &&
                        oldosc_x == origin_x &&
                        oldosc_y == origin_y &&
                        oldosc_z == origin_z &&
                        rstart.getTime() - last_render.getTime() < 100)
                    rapid_re_render_count++;
                else
                    rapid_re_render_count = 0;
            }

            if( rapid_re_render_count > 3 ) {
                rapid_re_render_count = 0;
                CheeseLab.dmesg( 2, "Rapid re-rendering with same origin detected, sleeping a bit to slow down");
                try {
                    Thread.sleep( 500 );
                } catch( InterruptedException e ) {}
            }

            boolean rv = false;
            /* Handle VolatileImage stuff */
            do {
                int valCode = osc.validate( gc );
                if (valCode == VolatileImage.IMAGE_INCOMPATIBLE) {
                    CheeseLab.dmesg( 2, "OSC recreated" );
                    // Create new OSC
                    synchronized( osclock ) {
                        /* Save width and height so we can release osc
                         * before we create a new one
                         */
                        int w = osc.getWidth();
                        int h = osc.getHeight();
                        
                        osc.flush();
                        osc = null;
                        osc = canvas.createVolatileImage( w, h );
                        canvas.setImageAccelerationPriority( osc, 0.8 );
                    }
                }
                
                rv = drawMap();
                
            } while( osc.contentsLost() && rv );
            
            if( rv ) {
                // Swap OSC
                synchronized(osclock) {
                    VolatileImage tmp = osc;
                    osc = oldosc;
                    oldosc = tmp;
                    oldosc_x = origin_x;
                    oldosc_y = origin_y;
                    oldosc_z = origin_z;
                }

                complete = true;
                
                // TODO: Investigate the possibility of a race condition here (between check and lock)
                if( waiting ) {
                    // Notify waiting thread that we are done
                    synchronized( this ) {
                        notifyAll();
                    }
                } else {
                    // Request a repaint if nobody is waiting anymore.
                    // Accelerated images don't like the notion of being
                    // touched by more than one thread at all, thus we have
                    // to make sure it will not happen and paint in draw_map
                    // only. That's why we request a full repaint from canvas.
                    skiprender=true;
                    canvas.map_changed();
                    canvas.changed();
                }
            } else {
                CheeseLab.dmesg( 2, "Rendering aborted due to new redraw!");
            }

            Date rend = new Date();
            long rdiff = rend.getTime() - rstart.getTime();
            double secs = (double)rdiff / 1000;
            CheeseLab.dmesg( 2, "Render time: " + rdiff + "ms (" + secs + " secs)" );

            // Update rendering statistics
            last_time = rdiff;
            if( rdiff < min_time ) min_time = rdiff;
            if( rdiff > max_time ) max_time = rdiff;
            // Calculate 20 sample moving average: more weight to recent draws
            if( avg_time == 0 )
                avg_time = rdiff;
            else
                avg_time = avg_time * 19/20 + rdiff * 1 / 20;

            /* NB! This line must be here and executed even if rendering
             * was aborted in order to detect rapid re-renders
             */
            last_render = rend;
        }
        CheeseLab.dmesg( 2, rendererThread.getName() + " stopped");
    }
    
    protected void clearImage( Image o ) {
        if( o == null )
            return; // Nothing to do

        clearImage(o, 0, 0, o.getWidth( null ), o.getHeight( null ) );
    }

    protected void clearImage( Image o, int x, int y, int w, int h ) {
        if( o == null )
            return; // Nothing to do

        Graphics g = o.getGraphics();
        
        g.setColor( Color.black );
        g.fillRect( x, y, w, h );
    }

    public void draw_map( Image buffer ) {
        Graphics g = buffer.getGraphics();
        
        // Only draw OSC if we just finished and this is just a repaint
        if( skiprender ) {
            skiprender = false;
            g.drawImage( oldosc, 0, 0, null );
            return;
        }

        // Make sure the map matches the size of it
        synchronized( osclock ) {
            //mapcache = buffer;
            if( osc != null ) {
                if( osc.getWidth() != buffer.getWidth( null ) ) {
                    osc.flush();
                    osc = null;
                } else if( osc.getHeight() != buffer.getHeight( null ) ) {
                    osc.flush();
                    osc = null;
                }
            }

            if( osc == null ) {
                CheeseLab.dmesg( 2, "Creating new image for off-screen rendering" );
                osc = canvas.createVolatileImage(
                          buffer.getWidth( null ), buffer.getHeight( null ) );
                canvas.setImageAccelerationPriority( osc, 0.8 );
            }
        }

        // Notify rendering thread that it should start it's work
        needtorender++;
        synchronized( this ) {
            notifyAll();
        }

        /* Wait for rendering to complete but not for more than 20ms.
         * The idea behind this arrangement is that user will probably not
         * notice a GUI "freeze" that lasts for less than 20ms and this way
         * we save some repaints for fast rendering cases and thus
         * reduce CPU load.
         */
        waiting = true;
        synchronized( this ) {
            try {
                wait( 20 );
            } catch( InterruptedException e ) {}
        }
        waiting = false;
        
        if( complete ) {
            drawOldOSC( g, 0, 0 );
            return;
        }

        /* If rendering didn't complete in time, we take an old osc
         * and draw it to new position (if possible).
         */
        synchronized( osclock ) {
            if( oldosc != null && origin_z == oldosc_z ) {
                int offset_x = 0, offset_y = 0;

                Point p = canvas.global_to_local( oldosc_x, oldosc_y, oldosc_z );
                offset_x = p.x;
                offset_y = p.y;

                CheeseLab.dmesg( 3, "Drawing old screen to offset: " + offset_x + "," + offset_y);
                drawOldOSC( g, offset_x, offset_y );

                // Fill the rest with black
                g.setColor( Color.black );
                if( offset_x > 0)
                    g.fillRect( 0, 0, offset_x, buffer.getHeight( null ) );
                else
                    g.fillRect( buffer.getWidth( null ) + offset_x, 0, -offset_x, buffer.getHeight( null ) );
                if( offset_y > 0)
                    g.fillRect( 0, 0, buffer.getWidth( null ), offset_y );
                else
                    g.fillRect( 0, buffer.getHeight( null ) + offset_y, buffer.getWidth( null ), -offset_y );
            } else {
                clearImage( buffer );
            }
        }

        drawLoadingText( buffer );
    }

    private void drawLoadingText( Image buffer ) {
        Graphics g = buffer.getGraphics();
        
        // Draw visual indication
        String str = "Loading ... ";
        FontMetrics metrics = g.getFontMetrics();
        Rectangle2D bounds = metrics.getStringBounds( str, g );
        g.setColor( Color.LIGHT_GRAY );
        g.drawString( str,
                      (int)(buffer.getWidth( null ) - bounds.getWidth()) + 1,
                      (int)(buffer.getHeight( null ) - bounds.getHeight()) + 1 );
    }

    public boolean prepare_move(Integer [] rooms) {
        int rows;
        int cols;
        
        // Construct an area which corresponds to the rooms we want to move
        move_area = new Area();

        move_rooms.clear();
        move_affected_rooms.clear();
        move_hidden_rooms.clear();
        
        if( rooms == null || rooms.length <= 0 )
            return false;

        Dimension tile_size = canvas.get_tile_size();
        
        /* Calculate areas and other needed info */
        for( int i = 0; i < rooms.length; i++ ) {
            CheeseRoomData room = CheeseMap.map.get_room( rooms[i].intValue(), 1 );
            move_rooms.put( new Integer( room.roomid ), room );
            /* Only create areas for rooms on other z-planes if they are visible
             * If they can become visible during the move operation, add them
             * to move_hidden_rooms
             */
            if( room.z != origin_z ) {
                if( room.z == origin_z + 1 ) {
                    if( CheeseMap.map.get_roomid_by_coord( room.x, room.y, origin_z ) != null ) {
                        move_hidden_rooms.add( room );
                        continue;
                    }
                }
                else
                    continue;
            }else{
                /* Add to hidden rooms also the rooms that are stationary, but might get
                 * revealed because of the move
                 */
                CheeseRoomData r1 = CheeseMap.map.get_room( room.x, room.y, room.z + 1 );
                if( r1 != null )
                    move_hidden_rooms.add( r1 );
            }

            Point p = canvas.global_to_local( room.x, room.y, room.z );
            move_area.add( new Area( new Rectangle( p, tile_size ) ) );
        }
        
        /* That's it for small tile sizes, as we don't draw warplines
         * we don't care about affected rooms and other such stuff
         */
        if( tile_width < no_overlays_thr ) {
            return true;
        }
        
        /* Create a vector of visible rooms that might develop warplines
         * to one of the moved rooms
         */
        synchronized(osclock) {
            if( osc == null )
                return false; // Nothing to do (yet)

            cols = osc.getWidth() / tile_width;
            rows = osc.getHeight() / tile_height;
        }
        
        for( int x = 0; x <= cols; x++ ) {
            for( int y = 0; y <= rows; y++ ) {
                CheeseRoomData room = CheeseMap.map.get_room( origin_x + x,
                        origin_y + y, origin_z );
                
                if( room == null )
                    continue;
                
                if( move_rooms.get( new Integer( room.exit_n ) ) != null ||
                    move_rooms.get( new Integer( room.exit_s ) ) != null ||
                    move_rooms.get( new Integer( room.exit_e ) ) != null ||
                    move_rooms.get( new Integer( room.exit_w ) ) != null ||
                    move_rooms.get( new Integer( room.exit_u ) ) != null ||
                    move_rooms.get( new Integer( room.exit_d ) ) != null )
                        move_affected_rooms.put( new Integer( room.roomid ), room );
            }
        }
        
        move_affected_rooms.putAll( move_rooms );
        
        move_repaint = true;
        
        // Ask for re-render
        needtorender++;
        synchronized( this ) {
            notifyAll();
        }

        waiting = true;
        synchronized( this ) {
            try {
                wait( 100 );
            } catch( InterruptedException e ) {}
        }
        waiting = false;

        return true;
    }

    /* This function draws the picture when rooms are being dragged around with
     * the mouse. I would love to ask renderer to render a new bitmap each mouse
     * move, but it's just not going to work: at first it would require mangling
     * with room and coordinate cache in unpredictable ways and painting clips of
     * a pre-rendered bitmap is definately faster.
     * However the speed comes with the cost of complexity and a number of
     * in-memory lists, that increase memory usage.
     */
    public void draw_move(Image buffer, int dx, int dy) {
        Graphics2D g = null;
        Vector w = new Vector();

        try {
            g = (Graphics2D)buffer.getGraphics();
        } catch( ClassCastException e ) {
            // This code really needs a Graphics2D object
            return;
        }
        
        Dimension tile_size = canvas.get_tile_size();

        drawOldOSC( g, 0, 0 );
        
        if( !complete ) {
            drawLoadingText( buffer );
            return;
        }
        
        /* Manipulate room coordinates.
         * This is evil, because it changes the original object that
         * is present in memory cache, however to do it othervise we
         * would have to copy the whole get_exit_tile logic.
         * Also it's not as bad as it looks because disk cache can not
         * become corrupted no matter what happens to the in-memory room
         * cache (this is in stark contrast to coordinate cache that is
         * saved from memory to disk and will become corrupted if altered
         * in unpredictable ways, fortunately coordinate cache is not directly
         * available to other classes).
         * FIXME: Maybe it's not such a bad idea after all?
         */
        Iterator it = move_rooms.values().iterator();
        while( it.hasNext() ) {
            CheeseRoomData room = (CheeseRoomData)it.next();
            room.x += dx;
            room.y += dy;
        }

        // Fill the space where selected rooms used to be
        g.setClip( move_area );
        g.setColor( new Color( 50, 50, 50 ) );
        g.fillRect(0, 0, buffer.getWidth( null ), buffer.getHeight( null ) );
        g.setClip( null );
        
        // Draw "hidden" rooms
        for( int i = 0; i < move_hidden_rooms.size(); i++ ) {
            CheeseRoomData room = (CheeseRoomData)move_hidden_rooms.elementAt( i ); 
            CheeseRoomData r2 = CheeseMap.map.get_room( room.x, room.y, origin_z );
            
            // Skip if the room overhead is stationary and we are still not visible
            if( r2 != null && move_rooms.get( new Integer( r2.roomid ) ) == null )
                continue;

            if( room.z == origin_z + 1 ) {
                Point p = canvas.global_to_local( room.x, room.y, room.z );
                if( tile_width < no_overlays_thr ) {
                    String tilename = canvas.tilename( room );
                    g.drawImage( get_image( "t_" + tilename, render_grid ), p.x, p.y, null );
                }else{
                    RenderCacheItem r = get_cached_tile( room );
                    g.drawImage( r.image, p.x, p.y, null );
                }
                g.drawImage( get_image( "overlay_blacken" ), p.x, p.y, null );
            }
        }

        // Transform the clip to new position
        AffineTransform move_t = new AffineTransform();
        move_t.translate( dx * tile_size.width, dy * tile_size.height );
        
        // Draw rooms to new position
        g.setClip( move_area.createTransformedArea( move_t ) );
        drawOldOSC( g, dx * tile_size.width, dy * tile_size.height );
        g.setClip( null );

        // Calculate warp lines
        if( tile_width >= no_overlays_thr ) {
            it = move_affected_rooms.values().iterator();
            while( it.hasNext() ) {
                CheeseRoomData room = (CheeseRoomData)it.next();
                
                get_exit_tile( room, room.exit_n, "", "n", w );
                get_exit_tile( room, room.exit_s, "", "s", w );
                get_exit_tile( room, room.exit_e, "", "e", w );
                get_exit_tile( room, room.exit_w, "", "w", w );
                get_exit_tile( room, room.exit_u, "", "u", w );
                get_exit_tile( room, room.exit_d, "", "d", w );
            }
            
            // Add other warplines as well
            w.addAll( warplines );
            
            // draw warp lines
            g.setColor( new Color( 255, 0, 0 ) );
            for( int i = 0; i < w.size(); i++ ) {
                int[] points = (int[])w.get( i );

                // only on this zplane
                if( points[ 2 ] == origin_z ) {
                    Point pstart = canvas.global_to_local_reference( points[ 0 ], points[ 1 ], points[ 2 ],
                                   origin_x, origin_y, origin_z );
                    Point pend = canvas.global_to_local_reference( points[ 3 ], points[ 4 ], points[ 5 ],
                                 origin_x, origin_y, origin_z );
    
    
                    g.drawLine( pstart.x + points [ 6 ], pstart.y + points[ 7 ],
                                 pend.x + tile_width / 2, pend.y + tile_height / 2 );
                }
            }
        }
        
        // Restore room coordinates
        it = move_rooms.values().iterator();
        while( it.hasNext() ) {
            CheeseRoomData room = (CheeseRoomData)it.next();
            room.x -= dx;
            room.y -= dy;
        }

        g.setColor( Color.lightGray );
        g.drawString( "Move: " + dx + "," + dy, 5, 20 );
    }


    // This function draws just the map, not player tiles or HUD data
    // which may change without the map data moving
    protected boolean drawMap() {
        rendering = needtorender;
        int cols, rows, width, height;
        Graphics2D sg;

        synchronized(osclock) {
            if( osc == null )
                return false; // Nothing to do (yet)

            cols = osc.getWidth() / tile_width;
            rows = osc.getHeight() / tile_height;
            sg = (Graphics2D)osc.getGraphics();
            width = osc.getWidth();
            height = osc.getHeight();
        }

        CheeseLab.dmesg( 3, "Drawing!" );

        // We mark the sync now so that if there are changed while we
        // are rendering, we will get them
        int rend_x = origin_x,
            rend_y = origin_y,
            rend_z = origin_z;

        CheeseLab.dmesg( 3, "Size: " + rows + "," + cols );
        CheeseLab.dmesg( 3, "Origin: " + rend_x + "," + rend_y + "," + rend_z );

        // Figure our bounding box
        int start_x = rend_x;
        int start_y = rend_y;
        int start_z = rend_z;

        int end_x = start_x + cols;
        int end_y = start_y + rows;
        int end_z = rend_z + 1;

        warplines.clear();

        // Make a slightly larger font
        sg.setFont( new Font( null, Font.BOLD, 12 ) );

        sg.setColor( new Color( 50, 50, 50 ) );
        sg.fillRect( 0, 0, width, height );

        Vector textnodes = new Vector();
        int x, y, z, local_x, local_y;

        for( z = 1; z >= 0; z-- ) {
            for( x = 0; x <= cols; x++ ) {
                local_x = x * tile_width;
                for( y = 0; y <= rows; y++ ) {

                    // If needtorender changes, it means we are lagging behind
                    if (rendering != needtorender)
                        return false;

                    local_y = y * tile_height;
                    int global_x = start_x + x;
                    int global_y = start_y + y;
                    int global_z = start_z + z;

                    CheeseRoomData room = CheeseMap.map.get_room( global_x, global_y, global_z );
                    if( room != null ) {
                        RenderCacheItem thisroom = null;

                        /* Skip overlays and warplines if tile size is too small.
                         * Overlays will not be visible below a certain threshold anyway,
                         * and warplines clutter the view on very small tile sizes.
                         * Also as no overlays will be drawn, there is no point to cache
                         * tiles as well, which will decrease memory usage.
                         */
                        if( tile_width < no_overlays_thr ) {
                            String tilename = canvas.tilename( room );
                            sg.drawImage( get_image( "t_" + tilename, render_grid ), local_x, local_y, null );
                        }else{
                            thisroom = get_cached_tile( room );
                            sg.drawImage( thisroom.image, local_x, local_y, null );
                            // Don't draw warplines for rooms that are going to be moved around
                            if( ! move_repaint ||
                                move_affected_rooms.get( new Integer( room.roomid ) ) == null )
                                warplines.addAll( thisroom.warplines );
                        }

                        if( global_z == rend_z ) {
                            // render notes text
                            if( room.notes != null && (! room.notes.equals( "" ) ) ) {
                                if( room.notes.length() < canvas.get_render_notes_length() && canvas.get_render_notes() ) {
                                    textnodes.add( room );
                                } else {
                                    if( thisroom != null && ! thisroom.drew_an_overlay ) {
                                        // Render a note tile if notes are here
                                        CheeseLab.dmesg( 3, "Doing default notes overlay" );
                                        sg.drawImage( get_image( "overlay_notes" ), local_x, local_y, null );
                                    }
                                }
                            }

                        } else {
                            sg.drawImage( get_image( "overlay_blacken" ), local_x, local_y, null );
                        }

                        // Selected
                        /*if( canvas.room_is_selected( room.roomid ) ) {
                            CheeseLab.dmesg( 3, "Drawing selected" );
                            sg.drawImage( get_image( "selected" ), local_x, local_y, null );
                        }*/

                        if( room.roomid == CheeseMap.edit_roomid )
                            sg.drawImage( get_image( "infobox" ), local_x, local_y, null );

                    }
                }
            }
        }
        
        // Don't draw warplines or notes when we are about to move rooms around
        if( move_repaint ) {
            move_repaint = false;
            sg.dispose();
            return true;
        }
        
        // warp lines
        sg.setColor( new Color( 255, 0, 0 ) );
        for( int i = 0; i < warplines.size(); i++ ) {
            // If needtorender changes, it means we are lagging behind
            if (rendering != needtorender)
                return false;

            int[] points = (int[])warplines.get( i );
            // only on this zplane
            if( points[ 2 ] == rend_z ) {
                Point pstart = canvas.global_to_local_reference( points[ 0 ], points[ 1 ], points[ 2 ],
                               rend_x, rend_y, rend_z );
                Point pend = canvas.global_to_local_reference( points[ 3 ], points[ 4 ], points[ 5 ],
                             rend_x, rend_y, rend_z );


                sg.drawLine( pstart.x + points [ 6 ], pstart.y + points[ 7 ],
                             pend.x + tile_width / 2, pend.y + tile_height / 2 );
            }
        }

        // Notes
        for( int i = 0; i < textnodes.size(); i++ ) {
            // If needtorender changes, it means we are lagging behind
            if (rendering != needtorender)
                return false;

            CheeseRoomData room =(CheeseRoomData)textnodes.get( i );
            Point p = canvas.global_to_local_reference( room.x, room.y, room.z, rend_x, rend_y, rend_z );
            local_x = p.x;
            local_y = p.y;

            int offset_x = 4;
            int offset_y = tile_height - 4;

            if( room.x % 2 == 0 ) {
                offset_y = tile_height / 2;
            }

            sg.setColor( Color.BLACK );
            sg.drawString( room.notes,
                           local_x + offset_x, local_y + offset_y );
            sg.setColor( Color.LIGHT_GRAY );
            sg.drawString( room.notes,
                           local_x + offset_x + 1, local_y + offset_y + 1 );
        }

        sg.dispose();

        return true;
    }
    
    /**
     * This function will draw old osc to Graphics object validating it properly 
     */
    private void drawOldOSC( Graphics g, int x, int y ) {
        if( oldosc == null )
            return;
        
        boolean lost = false;
        
        int valCode = oldosc.validate( gc );
        if( valCode != VolatileImage.IMAGE_OK ) {
            lost = true;
        }else{
            g.drawImage( oldosc, x, y, null );
            
            if( oldosc.contentsLost() )
                lost = true;
        }

        if( lost ) {
            CheeseLab.dmesg( 2, "Surface loss on OSC detected" );
            
            /* If we are currently rendering we don't care because 
             * next redraw will make things groovy again, othervise request
             * repaint from canvas
             */
            if( complete ) {
                canvas.map_changed();
                canvas.changed();
            }
        }
    }
    
    public void set_canvas(CheeseCanvas c) {
        if( canvas != null )
            canvas.removeCanvasListener( this );
        
        super.set_canvas( c );
        
        canvas.addCanvasListener( this );
    }
    
    // Kill ourselves!
    public void dispose() {
        stopThread = true;
        needtorender++;

        synchronized( this ) {
            notifyAll();
        }
        
        if( canvas != null )
            canvas.removeCanvasListener( this );
        
        CheeseMap.removeCheeseMapListener( this );
        super.dispose();
    }
    
    public static String get_sysinfo() {
        if( min_time == Long.MAX_VALUE ) // No statistics available yet
            return "";

        return "Rendering times:\n" +
            "    Last/Avg: " + last_time  + "ms / " + avg_time + "ms\n" +
            "    Min/Max: " + min_time + "ms / " + max_time + "ms\n";
    }

    public void roomChanged(CheeseRoomData newroom, CheeseRoomData oldroom) {
        skiprender = false;
    }

    public void roomDeleted(CheeseRoomData room) {
        skiprender = false;
    }

    public void tileChanged(String fname) {
        skiprender = false;
    }
    
    public void viewChanged(CheeseCanvas c) {
        skiprender = false;
    }

    public void refreshStarted() { }
    public void refreshFinished() { }
    public void onlineChanged(boolean online) { }
    public void playerChanged(CheesePlayerData player, CheesePlayerData oldplayer) { }
}
